#include "mainwindow.h"
#include "server.h"
#include "settings.h"
#include "settingsmenu.h"
#include "messages.h"

#include <gui/requesters.h>
#include <gui/spinner.h>
#include <gui/stringview.h>

using namespace os;

MainWindow::MainWindow() : os::Window( os::Rect( 0, 0, 700, 300 ), "main_wnd", "FTP Client" )
{
	m_pcSettings = new FTPSettings();
	m_pcSettings->Load();		/* Load settings from the default file */

	SetupMenus();
	SetupToolBar();
	
	SetupStatusBar();
	Layout();	

	/* Set Icon */
	os::Resources cCol( get_image_id() );
	os::ResStream *pcStream = cCol.GetResourceStream( "icon48x48.png" );
	os::BitmapImage *pcIcon = new os::BitmapImage( pcStream );
	SetIcon( pcIcon->LockBitmap() );
	delete( pcIcon );
	
	m_pcServer = NULL;
}

MainWindow::~MainWindow()
{
	if( m_pcServer ) m_pcServer->Close();
}

void MainWindow::SetupMenus()
{
	m_pcMainMenu = new os::Menu(os::Rect(0,0,1,1),"main_menu", os::ITEMS_IN_ROW);
	
	os::Menu* file = new os::Menu(os::Rect(0,0,1,1),"File",os::ITEMS_IN_COLUMN);
	file->AddItem("About",new os::Message(M_APP_ABOUT));
	file->AddItem(new os::MenuSeparator());
	file->AddItem("Quit",new os::Message(M_APP_QUIT));
	
	
	os::Menu* edit = new os::Menu(os::Rect(0,0,1,1),"Edit",os::ITEMS_IN_COLUMN);

	m_pcSettingsMenu = new SettingsMenu( "Settings", m_pcSettings );

	os::Menu* help = new os::Menu(os::Rect(0,0,1,1),"Help",os::ITEMS_IN_COLUMN);

	m_pcMainMenu->AddItem(file);
	m_pcMainMenu->AddItem(edit);
	m_pcMainMenu->AddItem(m_pcSettingsMenu);
	m_pcMainMenu->AddItem(help);
		
	
	m_pcMainMenu->SetTargetForItems(this);
	m_pcMainMenu->ResizeTo(os::Point(GetBounds().Width(),m_pcMainMenu->GetPreferredSize(false).y));
	AddChild(m_pcMainMenu);
}

void MainWindow::SetupToolBar()
{
	m_pcToolBar = new os::ToolBar(os::Rect(),"toolbar");
	m_pcToolBar->AddChild(m_pcConnectBar = new ConnectBar(),os::ToolBar::TB_FIXED_WIDTH);
	m_pcToolBar->SetFrame( os::Rect(0,m_pcMainMenu->GetBounds().Height()+1,GetBounds().Width(),m_pcMainMenu->GetBounds().Height()+1+m_pcToolBar->GetPreferredSize(false).y));
	AddChild(m_pcToolBar);
}

void MainWindow::SetupStatusBar()
{
	m_pcStatusBar = new os::StatusBar(os::Rect(0,GetBounds().Height()-20,GetBounds().Width(),GetBounds().Height()),"",os::CF_FOLLOW_LEFT | os::CF_FOLLOW_RIGHT | os::CF_FOLLOW_BOTTOM);
	m_pcStatusBar->AddPanel("status","status");
	AddChild(m_pcStatusBar);
}

void MainWindow::Layout()
{
	m_pcView = new FTPView(os::Rect(0,(int)m_pcMainMenu->GetBounds().Height() + (int)m_pcToolBar->GetBounds().Height() + 1,GetBounds().Width(),GetBounds().Height()-20),"container_view");
	AddChild(m_pcView);
}


String MainWindow::ConstructURL( const String& zHost, int nPort, const String& zUser, const String& zPassword )
{
	/* TODO: Lots of sanitizing, escaping and sanity checking */
	String zURL = "ftp://";
	if( zUser != "" ) {
		zURL += zUser;
		zURL += ":";
		zURL += zPassword;
		zURL += "@";
	}
	zURL += zHost;
	if( nPort > 0 ) {
		String zTmp;
		zTmp.Format( ":%i", nPort );
		zURL += zTmp;
	}
	return( zURL );
}


void MainWindow::HandleMessage( os::Message* pcMessage )
{
	switch( pcMessage->GetCode() )
	{
		case M_APP_ABOUT:
		{
			break;
		}
		
		case M_APP_QUIT:
		{
			PostMessage( os::M_QUIT );
			break;
		}
		
		case M_CONNECT:
		{
			String zHost = m_pcConnectBar->GetHost();
			int nPort = m_pcConnectBar->GetPort();
			String zUser = m_pcConnectBar->GetUser();
			String zPassword = m_pcConnectBar->GetPassword();
			String zURL = ConstructURL( zHost, nPort, zUser, zPassword );

			printf( "Creating connection to %s\n", zURL.c_str() );
			Server* pcServer = new FTPServer( zURL, this, 2 );
			if( !pcServer ) {
				printf( "FTP Client: Could not create Server( %s )!\n", zURL.c_str() );
				break;
			}
			/* TODO: check that the new url is actually different to the old one, and only proceed if so */
			m_pcView->SetServer( pcServer );
			
			if( m_pcServer ) {
				m_pcServer->Close();
				/* We can forget about the old Server now - its transfer thread will clean up and delete itself. */
			}
			m_pcServer = pcServer;

			break;
		}
		case M_SETTINGS_PASSIVE:
		{
			bool bPassive;
			if( pcMessage->FindBool( "status", &bPassive ) == 0 ) {
				m_pcSettings->SetPassive( bPassive );
				printf( "SETTINGS_PASSIVE %s\n", bPassive?"true":"false" );
			}
			break;
		}
		case M_SETTINGS_SAVE_HISTORY:
		{
			bool bSave;
			if( pcMessage->FindBool( "status", &bSave ) == 0 ){
				m_pcSettings->SetSaveHistory( bSave );
				printf( "SETTINGSSAVE_HISTORY %s\n", bSave?"true":"false" );
			}
			break;
		}
		case M_SETTINGS_MAX_CONNECTIONS_MENU:
		{
			LayoutView* pcView = new LayoutView( Rect( 0,0,100,50 ), "max_connections_layout" );
			HLayoutNode* pcNode = new HLayoutNode( "max_connections_node" );
			pcNode->AddChild( new StringView( Rect( 0,0,80,50 ), "max_connections_label", "Maximum simultaneous connections:" ) );
			Spinner* pcSpinner = new Spinner( Rect( 81,0,100,50 ), "max_connections_spinner", m_pcSettings->GetMaxConnections(), new Message( M_SETTINGS_MAX_CONNECTIONS ) );
			pcSpinner->SetMinMax( 1, 100 );
			pcSpinner->SetStep( 1 );
			pcSpinner->SetFormat( "%.0f" );
			pcSpinner->EnableStatusChanged( true );
			pcSpinner->SetTarget( this );
			pcNode->AddChild( pcSpinner );
			pcView->SetRoot( pcNode );
			
			Alert* pcAlert = new Alert( "Maximum simultaneous connections", pcView );
//			pcAlert->ResizeTo( pcView->GetPreferredSize( false ) );		/* Spinner doesn't give an appropriate size? */
			pcAlert->ResizeTo( pcView->GetPreferredSize( false ) + Point( 30, 12) );
			pcAlert->CenterInWindow( this );
			pcAlert->Show();
			pcAlert->MakeFocus( true );
			break;
		}
		case M_SETTINGS_MAX_CONNECTIONS:
		{
			Variant cValue;
			if( pcMessage->FindVariant( "value", &cValue ) == 0 ) {
				int nMax = cValue.AsInt32();
				if( nMax > 0 ) m_pcSettings->SetMaxConnections( nMax );
				printf( "SETTINGS_MAX_CONNECTIONS %i\n", nMax );
				m_pcSettingsMenu->SetSettings( m_pcSettings );
			} else {
				/* Libsyllable 0.6.5 Spinner doesn't include its current value in the status changed message :( */
				printf( "User changed the max connections setting, but because of a libsyllable bug we don't know what the new value is, sorry!\n" );
			}
			break;
		}
	}
}
bool MainWindow::OkToQuit()
{
	os::Application::GetInstance()->PostMessage( os::M_QUIT );
	return( true );
}







