#include <curl/curl.h>
#include <storage/file.h>
#include <util/messenger.h>
#include <util/datetime.h>

#include "queuenode.h"
#include "server.h"
#include "remotenode.h"
#include "messages.h"

extern "C" {
#include "ftpparse.h"
};

QueueNode::QueueNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nType, int nID )
{
	m_pcServer = pcServer;
	
	m_zLocalPath = zLocalPath;
	m_zRemotePath = zRemotePath;
	m_nType = nType;
	m_nStatus = NODE_QUEUED;
	m_nID = nID;
}

/* Set the necessary options in the curl handle. */
/* The important parts are done in the subclasses. */
CURLcode QueueNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	/* Call curl_easy_setopt() to set url, callbacks, etc. */
	/* We set some general options here; specifics like resume position etc will be set by subclasses. */

	/* Set callback functions */
	curl_easy_setopt( pHandle, CURLOPT_WRITEFUNCTION, QueueNode::WriteCallback );
	curl_easy_setopt( pHandle, CURLOPT_WRITEDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_READFUNCTION, QueueNode::ReadCallback );
	curl_easy_setopt( pHandle, CURLOPT_READDATA, this );
//	curl_easy_setopt( pHandle, CURLOPT_SEEKFUNCTION, QueueNode::SeekCallback );	/* Not available in older versions of libcurl */
//	curl_easy_setopt( pHandle, CURLOPT_SEEKDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSFUNCTION, QueueNode::ProgressBarCallback );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 0 );
	
	return( CURLE_OK );
}

QueueNode::~QueueNode()
{
}


/** The libcurl callback functions. Can't pass a virtual function to libcurl so we use these static wrappers. **/
size_t QueueNode::ReadCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Read( pBuf, nSize * nMult ) );
}

size_t QueueNode::WriteCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Write( pBuf, nSize * nMult ) );
}

int QueueNode::SeekCallback( void* pCookie, curl_off_t nOffset, int nOrigin )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Seek( nOffset, nOrigin ) );
}

int QueueNode::ProgressBarCallback( void* pCookie, double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->ProgressBar( fDownTotal, fDownNow, fUpTotal, fUpNow ) );
}

/**** DownloadNode class ****/
DownloadNode::DownloadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID )
	: QueueNode( pcServer, zLocalPath, zRemotePath, NODE_DOWNLOAD, nID )
{
	m_pcFile = NULL;	/* This is only created when SetupCurlHandleForTransfer() is first called. */
}

DownloadNode::~DownloadNode()
{
	if( m_pcFile ) delete( m_pcFile );
}

CURLcode DownloadNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_WRONLY | O_CREAT | O_TRUNC );
	}
	
	QueueNode::SetupCurlHandleForTransfer( pHandle );
	

	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	
	/* Set the url, set to download */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	printf( "DownloadNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 0 );
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
	
	return( CURLE_OK );
}

size_t DownloadNode::Write( void* pBuf, size_t nSize )
{
/*	char zTmp[1025];
	memcpy( zTmp, pBuf, 1024 );
	zTmp[1024] = 0;
	printf( "DownloadNode::Write( %i bytes ):\n%s\n", nSize, zTmp );
*/
//	printf( "DownloadNode::Write( %i bytes ) to %s\n", nSize, m_zLocalPath.c_str() );
	size_t nResult = m_pcFile->Write( pBuf, nSize );
	m_pcFile->Flush();
	return( nResult );
}

int DownloadNode::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );
	/* TODO: make sure that File::Seek() returns the same as libcurl expects */
}

int DownloadNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	printf( "Progress for download %s: %.1f%%\n", m_zRemotePath.c_str(), fDownNow/fDownTotal * 100 );
	return( 0 );
}

/**** UploadNode class ****/
UploadNode::UploadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID )
	: QueueNode( pcServer, zLocalPath, zRemotePath, NODE_UPLOAD, nID )
{
	m_pcFile = NULL;	/* This is only created when SetupCurlHandleForTransfer() is first called. */
}

UploadNode::~UploadNode()
{
	if( m_pcFile ) delete( m_pcFile );
}

CURLcode UploadNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_RDONLY );
	}
	
	QueueNode::SetupCurlHandleForTransfer( pHandle );
	

#if 1
	/* I don't think resuming uploads is this easy with ilbcurl :( */
	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
#else
	m_pcFile->Seek( 0, SEEK_SET );
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
#endif

	curl_off_t nSize = m_pcFile->GetSize();
	curl_easy_setopt( pHandle, CURLOPT_INFILESIZE_LARGE, nSize );
	
	/* Set the url, set to upload */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	printf( "UploadNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 1 );
	
	return( CURLE_OK );
}

size_t UploadNode::Read( void* pBuf, size_t nSize )
{
//	printf( "UploadNode::Read( %i )\n", nSize );
	return( m_pcFile->Read( pBuf, nSize ) );
}

int UploadNode::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );
	/* TODO: make sure that File::Seek() returns the same as libcurl expects */
}

int UploadNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	printf( "Progress for upload %s: %.1f%%\n", m_zRemotePath.c_str(), (float)(100 * fUpNow/fUpTotal) );
	return( 0 );
}

/**** DirListNode class ****/
DirListNode::DirListNode( Server* pcServer, const String& zRemotePath, Handler* pcTarget, int nID )
	: QueueNode( pcServer, "", zRemotePath, NODE_DIRLIST, nID )
{
	m_pcTarget = pcTarget;
	m_bInitial = true;
}

DirListNode::~DirListNode()
{
}

CURLcode DirListNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	
	QueueNode::SetupCurlHandleForTransfer( pHandle );
	
	String zURL = zServer;
	zURL += "/";
	zURL += m_zRemotePath;
	zURL += "/";			/* Make it clear to libcurl that we want a dir listing */
	

	/* Set the url, set to upload */
	printf( "DirListNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 0 );
//	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	
	return( CURLE_OK );
}

/* strnchr: find the first occurrence of nChar in the first nLength bytes of pzStr. If not found, points to the first char *after* the end of the string. */
char* strnchr( char* pzStr, size_t nLength, int nChar )
{
	size_t nPos = 0;
	while( nPos < nLength && pzStr[nPos] != nChar && pzStr[nPos] != 0 ) ++nPos;
	return( pzStr + nPos );
}

/* The dirlist parsing might be better in Server, but we need to do it here because we need to save any fragments of the last line. */
size_t DirListNode::Write( void* pBuf, size_t nSize )
{
	char* pzBuf = (char*)pBuf;
	String zTmp( pzBuf, nSize );
	printf( "DirListNode::Write( %s, %i bytes, %sinitial ):\n%s\n------ end listing data ------\n", m_zRemotePath.c_str(), (int)nSize, m_bInitial?"":"not ", zTmp.c_str() );

	std::vector< RemoteNode >* pacList = new std::vector< RemoteNode >;	/* Will be deleted by the FTPRemoteView when it's finished processing */

	/* Assuming (for now) that pzBuf ends with a \n at the end of a line. */
	struct ftpparse sData;
	memset( &sData, 0, sizeof( sData ) );
	
	size_t nTotalRead = 0;
	String zLine = m_zLastLineFragment;
	size_t nLineLength = m_zLastLineFragment.size();
	char* pzLine = pzBuf;
	nTotalRead = strnchr( pzLine, nSize, '\n' ) - pzLine;	/* Not counting the \n yet */
	nLineLength += nTotalRead;
	zLine += String( pzLine, nTotalRead );
	if( nTotalRead == nSize )	/* Didn't find a \n; the line is incomplete so save it for later */
	{
		m_zLastLineFragment = zLine;
		delete( pacList );
		return( nSize );
	}
	
	do {
//		printf( "pzBuf: %x  pzLine: %x  length: %i     pzLine - pzBuf: %i\n", pzBuf, pzLine, nLineLength, pzLine - pzBuf );
//		printf( "Line: '%s' length: %i\n", zLine.c_str(), nLineLength );

		int nResult;
		nResult = ftpparse( &sData, zLine.c_str(), nLineLength );
//		printf( "Parse result: %i\n", nResult );
		if( nResult == 1 ) {	/* 1 indicates success */
			RemoteNode cNode;
			cNode.m_zName = String( sData.name, sData.namelen );
			cNode.m_bIsDir = sData.flagtrycwd;
			cNode.m_nSize = sData.size;
			cNode.m_cTimestamp = DateTime( sData.mtime );
			cNode.m_zPath.Format( "%s/%s", m_zRemotePath.c_str(), cNode.m_zName.c_str() );
			/* ftpparse doesn't provide permissions! Guess we'll do without for now. */
			cNode.m_nPermissions = 0777;
//			printf( "Parsed: name '%s', path '%s', size %i, %s, %s\n", cNode.m_zName.c_str(), cNode.m_zPath.c_str(), cNode.m_nSize, (cNode.m_bIsDir?"dir":"file"), cNode.m_cTimestamp.GetDate().c_str() );
			pacList->push_back( cNode );
		} else {
			String zTmp = String( pzLine, nLineLength );	/* Just for the next printf */
			printf( "ParseDirListing: ftpparse() failed on '%s' (%i bytes)\n", zTmp.c_str(), nLineLength );
		}

		nTotalRead++;	/* Skip the \n */
		pzLine = pzBuf + nTotalRead;
		nLineLength = strnchr( pzLine, nSize - nTotalRead, '\n' ) - pzLine;
		nTotalRead += nLineLength;
		zLine = String( pzLine, nLineLength );
		if( nTotalRead >= nSize ) {
			/* Save anything after the last \n (this will be empty if the pzBuf ends with \n) */
			m_zLastLineFragment = zLine;
			break;
		}
	} while( 1 );
	
	printf( " nTotalRead: %i   nSize: %i   zLastLineFragment: '%s'\n", nTotalRead, nSize, m_zLastLineFragment.c_str() );
	
	if( pacList->size() == 0 ) {
		/* We didn't sucessfully parse any entries! Clean up. */
		delete( pacList );
		return( nSize );
	}

	Message cMsg( M_REMOTE_DIRLISTING );
	cMsg.AddString( "path", m_zRemotePath );
	cMsg.AddPointer( "list", pacList );
	if( m_bInitial ) cMsg.AddBool( "initial", true );	/* TODO: fix up initial, final flags. Maybe remove them. */
//	cMsg.AddBool( "final", true );
	
	Messenger cMessenger( m_pcTarget );
	cMessenger.SendMessage( &cMsg );
	
	m_bInitial = false;
	
	return( nSize );
}

int DirListNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
//	printf( "Progress for dirlisting %s: %.1f out of %.1f\n", m_zRemotePath.c_str(), (float)fDownNow, (float)fDownTotal );
	return( 0 );
}


