
#include "server.h"
#include "remotenode.h"
#include "messages.h"

Server::Server( const String& zURL, Handler* pcTarget, int nParallelConnections )
{
	m_zURL = zURL;
	m_pcTarget = pcTarget;
	m_nParallelConnections = nParallelConnections;
	
	m_nNextID = 1;
	
	m_pcLock = new Locker( "server_lock" );
	
	m_pcThread = new TransferThread( this );
	m_pcThread->Start();
}

Server::~Server()
{
	delete( m_pcLock );
	/* All other cleanup is done in TransferThread */
}

/* Must call Lock() before accessing the node queue. */
int Server::Lock()
{
	return( m_pcLock->Lock() );
}

int Server::Unlock()
{
	return( m_pcLock->Unlock() );
}

String Server::GetServerAddress()
{
	return( m_zURL );
}

int Server::Close()
{
	m_pcThread->SendMessage( M_TT_CLOSE );
	return( 0 );
}

/* Queue a download from the remote file to the local file. zRemoteSource is relative to the ftp server, eg /remotedir/file */
int Server::GetRemote( const String& zRemoteSource, const String& zLocalDest )
{
	printf( "GetRemote( %s, %s )\n", zRemoteSource.c_str(), zLocalDest.c_str() );
	Message* pcMsg = new Message( M_TT_ADD );
	pcMsg->AddString( "localpath", zLocalDest );
	pcMsg->AddString( "remotepath", zRemoteSource );
	pcMsg->AddInt32( "type", NODE_DOWNLOAD );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/* Queue an upload of the given local file to the given remote file. zRemoteDest is relative to the ftp server, eg /remotedir/file */
int Server::SendLocal( const String& zLocalSource, const String& zRemoteDest )
{
	printf( "SendLocal( %s, %s )\n", zLocalSource.c_str(), zRemoteDest.c_str() );
	Message* pcMsg = new Message( M_TT_ADD );
	pcMsg->AddString( "localpath", zLocalSource );
	pcMsg->AddString( "remotepath", zRemoteDest );
	pcMsg->AddInt32( "type", NODE_UPLOAD );
	return( m_pcThread->SendMessage( pcMsg ) );
}

int Server::GetDirListing( const String& zPath, Handler* pcTarget )
{
	Message* pcMsg = new Message( M_TT_ADD );
	pcMsg->AddString( "remotepath", zPath );
	pcMsg->AddString( "localpath", "" );
	pcMsg->AddInt32( "type", NODE_DIRLIST );
	pcMsg->AddPointer( "target", pcTarget );
	return( m_pcThread->SendMessage( pcMsg ) );
}

int Server::PauseTransfer( int nNodeID )
{
	Message* pcMsg = new Message( M_TT_PAUSE );
	pcMsg->AddInt32( "id", nNodeID );
	return( m_pcThread->SendMessage( pcMsg ) );
}

int Server::ResumeTransfer( int nNodeID )
{
	Message* pcMsg = new Message( M_TT_RESUME );
	pcMsg->AddInt32( "id", nNodeID );
	return( m_pcThread->SendMessage( pcMsg ) );
}

int Server::CancelTransfer( int nNodeID )
{
	Message* pcMsg = new Message( M_TT_CANCEL );
	pcMsg->AddInt32( "id", nNodeID );
	return( m_pcThread->SendMessage( pcMsg ) );
}

/**** The following methods are only called from within the transfer thread ****/

CURLcode Server::_SetupCurlHandle( CURL* pHandle )
{
	curl_easy_setopt( pHandle, CURLOPT_VERBOSE, 1 );	/* For testing */
//	curl_easy_setopt( pHandle, CURLOPT_FTPPORT, "-" );
	curl_easy_setopt( pHandle, CURLOPT_FTP_RESPONSE_TIMEOUT, (long)60 );	/* Supposedly in seconds!? But it must be huge or else everything times out! */
	curl_easy_setopt( pHandle, CURLOPT_TIMEOUT, (long)100000 );
	/* That's it for now; other options will be set by subclasses or by _SetupCurlHandleForTransfer() */
	return( CURLE_OK );
}

#if 0
int Server::ParseDirListing( const String& zDir, char* pzBuf, size_t nSize, Handler* pcTarget, bool bInitial )
{
	String zTmp( pzBuf, nSize );
	printf( "ParseDirListing( %s, %i bytes, %sinitial ):\n%s\n------ end listing ------\n", zDir.c_str(), (int)nSize, bInitial?"":"not ", zTmp.c_str() );

	std::vector< RemoteNode >* pacList = new std::vector< RemoteNode >;	/* Will be deleted by the FTPRemoteView when it's finished processing */

	/* Assuming (for now) that pzBuf ends with a \n at the end of a line. */
	struct ftpparse sData;
	memset( &sData, 0, sizeof( sData ) );
	
	char* pzLine = pzBuf;
	size_t nLineLength = strchr( pzLine, '\n' ) - pzLine;
	do {
		printf( "pzBuf: %x  pzLine: %x  length: %i     pzLine - pzBuf: %i\n", pzBuf, pzLine, nLineLength, pzLine - pzBuf );
		String zTmp = String( pzLine, nLineLength );	/* Just for the next printf */
		printf( "Line: '%s' length: %i\n", zTmp.c_str(), nLineLength );

		int nResult;
		nResult = ftpparse( &sData, pzLine, nLineLength );
		printf( "Parse result: %i\n", nResult );
		if( nResult == 1 ) {	/* 1 indicates success */
			RemoteNode cNode;
			cNode.m_zName = String( sData.name, sData.namelen );
			cNode.m_bIsDir = sData.flagtrycwd;
			cNode.m_nSize = sData.size;
			cNode.m_cTimestamp = DateTime( sData.mtime );
			cNode.m_zPath.Format( "%s/%s", zDir.c_str(), cNode.m_zName.c_str() );
			/* ftpparse doesn't provide permissions! Guess we'll do without for now. */
			cNode.m_nPermissions = 0777;
			printf( "Parsed: name '%s', path '%s', size %i, %s, %s\n", cNode.m_zName.c_str(), cNode.m_zPath.c_str(), cNode.m_nSize, (cNode.m_bIsDir?"dir":"file"), cNode.m_cTimestamp.GetDate().c_str() );
			pacList->push_back( cNode );
		} else {
			String zTmp = String( pzLine, nLineLength );	/* Just for the next printf */
			printf( "ParseDirListing: ftpparse() failed on '%s' (%i bytes)\n", zTmp.c_str(), nLineLength );
		}

		pzLine += nLineLength + 1;
		nLineLength = strchr( pzLine, '\n' ) - pzLine;
	} while( *pzLine != 0 );
	
	if( pacList->size() == 0 ) {
		/* We didn't sucessfully parse any entries! Clean up. */
		delete( pacList );
		return( nSize );
	}

	Message cMsg( M_REMOTE_DIRLISTING );
	cMsg.AddString( "path", zDir );
	cMsg.AddPointer( "list", pacList );
	if( bInitial ) cMsg.AddBool( "initial", true );	/* TODO: fix up initial, final flags. Maybe remove them. */
	cMsg.AddBool( "final", true );
	
	Messenger cMessenger( pcTarget );
	cMessenger.SendMessage( &cMsg );
	
	return( nSize );

#if 0
/*** For testing ***/
	if( zDir == "/" ) {
		std::vector< RemoteNode >* pacList = new std::vector< RemoteNode >;	/* Will be deleted by the FTPRemoteView when it's finished processing */
		RemoteNode cNode;
		cNode.m_zName = "banner";
		cNode.m_zPath = "banner";
		cNode.m_bIsDir = false;
		cNode.m_nSize = 3141;
		cNode.m_nPermissions = 0;
		cNode.m_cTimestamp = DateTime::Now();
		pacList->push_back( cNode );
		
		cNode.m_zName = "folder 1";
		cNode.m_bIsDir = true;
		cNode.m_nSize = 0;
		pacList->push_back( cNode );

		cNode.m_zName = "hello.txt";
		cNode.m_zPath = "/hello.txt";
		cNode.m_bIsDir = false;
		cNode.m_nSize = 2718;
		cNode.m_nPermissions = 0;
		cNode.m_cTimestamp = DateTime::Now();
		pacList->push_back( cNode );

		cNode.m_zName = "abc";
		cNode.m_bIsDir = true;
		cNode.m_nSize = 0;
		pacList->push_back( cNode );
		
		Message cMsg( M_REMOTE_DIRLISTING );
		cMsg.AddString( "path", zDir );
		cMsg.AddPointer( "list", pacList );
		cMsg.AddBool( "initial", true );
		cMsg.AddBool( "final", true );
		
		Messenger cMessenger( pcTarget );
		cMessenger.SendMessage( &cMsg );
	} else {
		std::vector< RemoteNode >* pacList = new std::vector< RemoteNode >;	/* Will be deleted by the FTPRemoteView when it's finished processing */
		RemoteNode cNode;
		cNode.m_zName = "z.txt";
		cNode.m_zPath = "/abc/z.txt";
		cNode.m_bIsDir = false;
		cNode.m_nSize = 2718;
		cNode.m_nPermissions = 0;
		cNode.m_cTimestamp = DateTime::Now();
		pacList->push_back( cNode );
		

		Message cMsg( M_REMOTE_DIRLISTING );
		cMsg.AddString( "path", zDir );
		cMsg.AddPointer( "list", pacList );
		cMsg.AddBool( "initial", true );
		cMsg.AddBool( "final", true );
		
		Messenger cMessenger( pcTarget );
		cMessenger.SendMessage( &cMsg );
	}
/*******************/

	return( nSize );
#endif
}
#endif

/**** FTPServer class (not implemented yet - all the functionality is in base Server class. Perhaps some functionality should be moved to FTPServer, or maybe we can delete FTPServer.) ****/

FTPServer::FTPServer( const String& zURL, Handler* pcTarget, int nParallelConnections )
	: Server( zURL, pcTarget, nParallelConnections )
{
}

FTPServer::~FTPServer()
{
}

